/*
 * kexec-elf32-ppc.c - kexec Elf32 loader for the PowerPC
 * Copyright (C) 2004 Albert Herranz
 *
 * This source code is licensed under the GNU General Public License,
 * Version 2.  See the file COPYING for more details.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>
#include <elf.h>
#include <boot/elf_boot.h>
#include <ip_checksum.h>
#include "kexec.h"
#include "kexec-ppc.h"

static int debug = 0;

#define BOOTLOADER         "kexec"
#define BOOTLOADER_VERSION VERSION
#define MAX_COMMAND_LINE   256

#define UPSZ(X) ((sizeof(X) + 3) & ~3)
static struct boot_notes {
	Elf_Bhdr hdr;
	Elf_Nhdr bl_hdr;
	unsigned char bl_desc[UPSZ(BOOTLOADER)];
	Elf_Nhdr blv_hdr;
	unsigned char blv_desc[UPSZ(BOOTLOADER_VERSION)];
	Elf_Nhdr cmd_hdr;
	unsigned char command_line[0];
} elf_boot_notes = {
	.hdr = {
		.b_signature = 0x0E1FB007,
		.b_size = sizeof(elf_boot_notes),
		.b_checksum = 0,
		.b_records = 3,
	},
	.bl_hdr = {
		.n_namesz = 0,
		.n_descsz = sizeof(BOOTLOADER),
		.n_type = EBN_BOOTLOADER_NAME,
	},
	.bl_desc = BOOTLOADER,
	.blv_hdr = {
		.n_namesz = 0,
		.n_descsz = sizeof(BOOTLOADER_VERSION),
		.n_type = EBN_BOOTLOADER_VERSION,
	},
	.blv_desc = BOOTLOADER_VERSION,
	.cmd_hdr = {
		.n_namesz = 0,
		.n_descsz = 0,
		.n_type = EBN_COMMAND_LINE,
	},
};


int elf32_ppc_probe(FILE * file)
{
	Elf32_Ehdr ehdr;
	if (fseek(file, 0, SEEK_SET) < 0) {
		fprintf(stderr, "seek error: %s\n", strerror(errno));
		return -1;
	}
	if (fread(&ehdr, sizeof(ehdr), 1, file) != 1) {
		if (debug) {
			fprintf(stderr, "File too short.\n");
		}
		return 0;
	}
	if (memcmp(ehdr.e_ident, ELFMAG, SELFMAG) != 0) {
		/* No ELF header */
		if (debug) {
			fprintf(stderr, "No ELF header.\n");
		}
		return 0;
	}
	if (ehdr.e_ident[EI_CLASS] != ELFCLASS32) {
		/* Not a 32bit ELF file */
		if (debug) {
			fprintf(stderr, "Not a 32bit ELF file.\n");
		}
		return 0;
	}
	if (ehdr.e_ident[EI_DATA] != ELFDATA2MSB) {
		/* not a big endian ELF file */
		if (debug) {
			fprintf(stderr, "Not a big endian ELF file.\n");
		}
		return 0;
	}
	if ((ehdr.e_ident[EI_VERSION] != EV_CURRENT) ||
	    (ehdr.e_version != EV_CURRENT)) {
		/* unkown elf version */
		if (debug) {
			fprintf(stderr, "Unknown ELF file version.\n");
		}
		return 0;
	}
	if (ehdr.e_type != ET_EXEC) {
		/* not an ELF executable */
		if (debug) {
			fprintf(stderr, "Not an ELF executable.\n");
		}
		return 0;
	}

	if (ehdr.e_ehsize != sizeof(Elf32_Ehdr)) {
		/* invalid ELF header size */
		if (debug) {
			fprintf(stderr, "Invalid ELF header size.\n");
		}
		return 0;
	}
	if (ehdr.e_phentsize != sizeof(Elf32_Phdr)) {
		/* invalid program header size */
		if (debug) {
			fprintf(stderr, "Invalid program header size.\n");
		}
		return 0;
	}
	if ((ehdr.e_phoff == 0) || (ehdr.e_phnum == 0)) {
		/* no program header */
		if (debug) {
			fprintf(stderr, "No program header.\n");
		}
		return 0;
	}
	/* Verify the architecuture specific bits */
	if (ehdr.e_machine != EM_PPC) {
		/* for a different architecture */
		if (debug) {
			fprintf(stderr, "Not for this architecture.\n");
		}
		return 0;
	}
	return 1;
}

void elf32_ppc_usage(void)
{
	printf
	    ("-d, --debug               Enable debugging to help spot a failure.\n"
	     "    --command-line=STRING Set the kernel command line to STRING.\n"
	     "    --append=STRING       Set the kernel command line to STRING.\n"
	     "    --gamecube=1|0        Enable/disable support for ELFs with changed\n"
	     "                          addresses suitable for the GameCube.\n");
}

int elf32_ppc_load(FILE * file, int argc, char **argv,
		   void **ret_entry, struct kexec_segment **ret_segments,
		   int *ret_nr_segments)
{
	struct kexec_segment *segment;
	int nr_segments;
	Elf32_Ehdr ehdr;
	Elf32_Phdr *phdr;
	size_t phdr_bytes;
	size_t arg_bytes;
	struct boot_notes *notes;
	size_t note_bytes;
	const char *command_line;
	int command_line_len;
	unsigned long mstart;
	unsigned char *setup_start;
	uint32_t setup_size;
	char *buf;
	size_t size;
#ifdef CONFIG_GAMECUBE
	int target_is_gamecube = 1;
#else
	int target_is_gamecube = 0;
#endif
	int i;
	int opt;
#define OPT_APPEND	(OPT_MAX+0)
#define OPT_GAMECUBE	(OPT_MAX+1)
	static const struct option options[] = {
		KEXEC_OPTIONS
		{"debug", 0, 0, OPT_DEBUG},
		{"command-line", 1, 0, OPT_APPEND},
		{"append", 1, 0, OPT_APPEND},
		{"gamecube", 1, 0, OPT_GAMECUBE},
		{0, 0, 0, 0},
	};

	static const char short_options[] = KEXEC_OPT_STR "d";

	debug = 0;
	command_line = 0;
	while ((opt = getopt_long(argc, argv, short_options, options, 0)) != -1) {
		switch (opt) {
		default:
			/* Ignore core options */
			if (opt < OPT_MAX) {
				break;
			}
		case '?':
			usage();
			return -1;
		case OPT_DEBUG:
			debug = 1;
			break;
		case OPT_APPEND:
			command_line = optarg;
			break;
		case OPT_GAMECUBE:
			target_is_gamecube = atoi(optarg);
			break;
		}
	}
	command_line_len = 0;
	if (command_line) {
		command_line_len = strlen(command_line) + 1;
	}

	/* Read in the Elf header */
	if (fseek(file, 0, SEEK_SET) != 0) {
		fprintf(stderr, "seek error: %s\n", strerror(errno));
		return -1;
	}
	if (fread(&ehdr, sizeof(ehdr), 1, file) != 1) {
		fprintf(stderr, "read error: %s\n", strerror(errno));
		return -1;
	}
	/* Read in the program header */
	phdr_bytes = sizeof(*phdr) * ehdr.e_phnum;
	phdr = malloc(phdr_bytes);
	if (phdr == 0) {
		fprintf(stderr, "malloc failed: %s\n", strerror(errno));
		return -1;
	}
	if (fseek(file, ehdr.e_phoff, SEEK_SET) != 0) {
		fprintf(stderr, "seek error: %s\n", strerror(errno));
		return -1;
	}
	if (fread(phdr, phdr_bytes, 1, file) != 1) {
		fprintf(stderr, "read error: %s\n", strerror(errno));
		return -1;
	}

	/* Setup the segments */
	segment = malloc(sizeof(*segment) * (ehdr.e_phnum + 1));
	if (segment == 0) {
		fprintf(stderr, "malloc failed: %s\n", strerror(errno));
		return -1;
	}

	/* Skip the argument segment */
	nr_segments = 0;
	segment[nr_segments].buf = 0;
	segment[nr_segments].bufsz = 0;
	segment[nr_segments].mem = 0;
	segment[nr_segments].memsz = 0;
	nr_segments++;

	/* Now all the rest of the segments */
	for (i = 0; i < ehdr.e_phnum; i++) {
		if (phdr[i].p_type != PT_LOAD) {
			continue;
		}
		size = phdr[i].p_filesz;
		if (size > phdr[i].p_memsz) {
			size = phdr[i].p_memsz;
		}
		buf = malloc(size);
		if (buf == 0) {
			fprintf(stderr, "malloc failed: %s\n", strerror(errno));
			return -1;
		}
		segment[nr_segments].buf = buf;
		segment[nr_segments].bufsz = size;
		mstart = phdr[i].p_paddr;
		if (target_is_gamecube) {
                        /*
                         * GameCube ELF kernel is linked with memory mapped
			 * this way (to easily transform it into a DOL
			 * suitable for being loaded with psoload):
                         *
                         * 80000000 - 817fffff 24MB RAM, cached
                         * c0000000 - c17fffff 24MB RAM, not cached
                         *
                         * kexec, instead, needs physical memory layout, so
                         * we clear the upper bits of the address.
                         * (2 bits should be enough, indeed)
                         */
			mstart &= ~0xf0000000;	/* clear bits 0-3, ibm syntax */
		}
		segment[nr_segments].mem = (void *)mstart;
		segment[nr_segments].memsz = phdr[i].p_memsz;
		if (valid_memory_range(segment + nr_segments) < 0) {
			fprintf(stderr, "Invalid memory segment %p - %p\n",
				segment[nr_segments].mem,
				((char *)segment[nr_segments].mem) +
				segment[nr_segments].memsz);
			return -1;
		}
		nr_segments++;
		if (size == 0) {
			/* Don't do file I/O if there is nothing in the file */
			continue;
		}
		if (fseek(file, phdr[i].p_offset, SEEK_SET) != 0) {
			fprintf(stderr, "seek failed: %s\n", strerror(errno));
			return -1;
		}
		if (fread(buf, size, 1, file) != 1) {
			fprintf(stderr, "Read failed: %s\n", strerror(errno));
			return -1;
		}
	}
	/* Generate and setup the argument segment */
	if (sort_segments(segment, nr_segments) < 0) {
		return -1;
	}
	if (target_is_gamecube) {
		setup_start = setup_dol_start;
		setup_size = setup_dol_size;
	} else {
		setup_start = setup_simple_start;
		setup_size = setup_simple_size;
	}
	note_bytes = sizeof(elf_boot_notes) + ((command_line_len + 3) & ~3);
	arg_bytes = note_bytes + ((setup_size + 3) & ~3);
	segment[0].bufsz = arg_bytes;
	segment[0].memsz = arg_bytes;
	if (locate_arguments(segment, nr_segments, 4, 0xFFFFFFFFUL) < 0) {
		return -1;
	}
	segment[0].buf = malloc(arg_bytes);
	if (segment[0].buf == 0) {
		fprintf(stderr, "malloc failed: %s\n", strerror(errno));
		return -1;
	}
	notes =
	    (struct boot_notes *)(((char *)segment[0].buf) +
				  ((setup_size + 3) & ~3));
	setup_dol_regs.spr8 = ehdr.e_entry;
	memcpy(segment[0].buf, setup_start, setup_size);
	memcpy(notes, &elf_boot_notes, sizeof(elf_boot_notes));
	memcpy(notes->command_line, command_line, command_line_len);
	notes->hdr.b_size = note_bytes;
	notes->cmd_hdr.n_descsz = command_line_len;
	notes->hdr.b_checksum = compute_ip_checksum(notes, note_bytes);

	*ret_entry = segment[0].mem;
	*ret_nr_segments = nr_segments;
	*ret_segments = segment;
	return 0;
}
